from __future__ import annotations

import enum
from multiprocessing import cpu_count
from pathlib import Path
from typing import Literal

import numpy as np
from numpy.random import SeedSequence
from numpy.typing import NDArray
from ruamel import yaml

_PKG_NAME: str = Path(__file__).parent.stem

VERSION = "2025.739290.3"

__version__ = VERSION

this_yaml = yaml.YAML(typ="safe", pure=True)
this_yaml.constructor.deep_construct = True
this_yaml.indent(mapping=2, sequence=4, offset=2)


DATA_DIR: Path = Path.home() / _PKG_NAME
"""
Defines a subdirectory named for this package in the user's home path.

If the subdirectory doesn't exist, it is created on package invocation.
"""
if not DATA_DIR.is_dir():
    DATA_DIR.mkdir(parents=False)

DEFAULT_REC_RATIO = 0.85

EMPTY_ARRAYDOUBLE = np.array([], float)
EMPTY_ARRAYINT = np.array([], int)

NTHREADS = 2 * cpu_count()

np.set_printoptions(precision=24, floatmode="fixed")

type HMGPubYear = Literal[1982, 1984, 1992, 2010, 2023]

type ArrayBoolean = NDArray[np.bool_]
type ArrayFloat = NDArray[np.float16 | np.float32 | np.float64 | np.float128]
type ArrayINT = NDArray[np.intp]

type ArrayDouble = NDArray[np.float64]
type ArrayBIGINT = NDArray[np.int64]

## Add yaml representer, constructor for NoneType
(_, _) = (
    this_yaml.representer.add_representer(
        type(None), lambda _r, _d: _r.represent_scalar("!None", "none")
    ),
    this_yaml.constructor.add_constructor("!None", lambda _c, _n, /: None),
)

## Add yaml representer, constructor for ndarray
(_, _) = (
    this_yaml.representer.add_representer(
        np.ndarray,
        lambda _r, _d: _r.represent_sequence("!ndarray", (_d.tolist(), _d.dtype.str)),
    ),
    this_yaml.constructor.add_constructor(
        "!ndarray", lambda _c, _n, /: np.array(*_c.construct_sequence(_n))
    ),
)

# Add yaml representer, constructor for SeedSequence
this_yaml.representer.add_representer(
    SeedSequence,
    lambda _r, _d: _r.represent_mapping(
        "!SeedSequence",
        {
            _a: getattr(_d, _a)
            for _a in ("entropy", "spawn_key", "pool_size", "n_children_spawned")
        },
    ),
)
this_yaml.constructor.add_constructor(
    "!SeedSequence", lambda _c, _n, /: SeedSequence(**_c.construct_mapping(_n))
)


@enum.unique
class RECForm(enum.StrEnum):
    """For derivation of recapture ratio from market shares."""

    INOUT = "inside-out"
    OUTIN = "outside-in"
    FIXED = "proportional"


@enum.unique
class UPPAggrSelector(enum.StrEnum):
    """
    Aggregator for GUPPI and diversion ratio estimates.

    """

    AVG = "average"
    CPA = "cross-product-share weighted average"
    CPD = "cross-product-share weighted distance"
    CPG = "cross-product-share weighted geometric mean"
    DIS = "symmetrically-weighted distance"
    GMN = "geometric mean"
    MAX = "max"
    MIN = "min"
    OSA = "own-share weighted average"
    OSD = "own-share weighted distance"
    OSG = "own-share weighted geometric mean"


for _typ in (RECForm, UPPAggrSelector):
    #  NOTE: If additional enums are defined in this module,
    #  add themn to the list above

    _, _ = (
        this_yaml.representer.add_representer(
            _typ,
            lambda _r, _d: _r.represent_scalar(f"!{_d.__class__.__name__}", _d.name),
        ),
        this_yaml.constructor.add_constructor(
            f"!{_typ.__name__}",
            lambda _c, _n, /: getattr(
                globals().get(_n.tag.lstrip("!")), _c.construct_scalar(_n)
            ),
        ),
    )
